@@ -15,12 +15,6 @@
 #ifndef RADCLI_H
 #define RADCLI_H
 
-#ifdef CP_DEBUG
-#define		DEBUG(args, ...)	rc_log(## args)
-#else
-#define		DEBUG(args, ...)	;
-#endif
-
 #include	<sys/types.h>
 /*
  * Include for C99 uintX_t defines is stdint.h on most systems.  Solaris uses
@@ -37,24 +31,20 @@
 #include <sys/socket.h>
 #include <netdb.h>
 
-#undef __BEGIN_DECLS
-#undef __END_DECLS
+/* *INDENT-OFF* */
 #ifdef __cplusplus
-# define __BEGIN_DECLS extern "C" {
-# define __END_DECLS }
-#else
-# define __BEGIN_DECLS /* empty */
-# define __END_DECLS /* empty */
+extern "C" {
 #endif
+/* *INDENT-ON* */
 
 /**
  * @defgroup radcli-api Main API
- * @brief Main API Functions 
+ * @brief Main API Functions
  *
  * @{
  */
 
-#define AUTH_PASS_LEN		(3 * 16) /* multiple of 16 */
+#define AUTH_PASS_LEN		(7 * 16) /* multiple of 16 */
 #define AUTH_ID_LEN		64
 
 #define BUFFER_LEN		8192
@@ -91,7 +81,7 @@ typedef struct rc_conf rc_handle;
  * Several of its fields have been deprecated.
  */
 typedef struct server {
-	int max; //!< for backwards compatibility; its value will be 1.
+	int   max;
 	char *name[SERVER_MAX];
 	uint16_t port[SERVER_MAX];
 	char *secret[SERVER_MAX];
@@ -103,7 +93,8 @@ typedef struct server {
 typedef enum rc_socket_type {
 	RC_SOCKET_UDP = 0,	//!< Plain UDP socket
 	RC_SOCKET_TLS = 1,	//!< TLS socket
-	RC_SOCKET_DTLS = 2	//!< DTLS socket
+	RC_SOCKET_DTLS = 2,	//!< DTLS socket
+	RC_SOCKET_TCP = 3	//!< Plain TCP socket
 } rc_socket_type;
 
 #define AUTH_HDR_LEN			20
@@ -392,6 +383,38 @@ typedef enum rc_acct_auth_type {
 	PW_REMOTE=3
 } rc_acct_auth_type;
 
+/** \enum rc_vendor_pec --- http://www.iana.org/assignments/enterprise-numbers/enterprise-numbers
+ */
+typedef enum rc_vendor_pec {
+  VENDOR_NONE=0,
+  VENDOR_MICROSOFT	     = 311,
+  VENDOR_ROARING_PENGUIN     = 10055
+} rc_vendor_type;
+
+/* Vendor RADIUS attribute-value pairs for MICROSOFT */
+enum rc_vendor_attr_microsoft {
+  PW_MS_CHAP_CHALLENGE	=	11,	/* string */
+  PW_MS_CHAP_RESPONSE	=	1,	/* string */
+  PW_MS_CHAP2_RESPONSE	=	25,	/* string */
+  PW_MS_CHAP2_SUCCESS	=	26,	/* string */
+  PW_MS_MPPE_ENCRYPTION_POLICY=	7,	/* string */
+  PW_MS_MPPE_ENCRYPTION_TYPE=	8,	/* string */
+  PW_MS_MPPE_ENCRYPTION_TYPES=PW_MS_MPPE_ENCRYPTION_TYPE,
+  PW_MS_CHAP_MPPE_KEYS	=	12,	/* string */
+  PW_MS_MPPE_SEND_KEY	=	16,	/* string */
+  PW_MS_MPPE_RECV_KEY	=	17,	/* string */
+  PW_MS_PRIMARY_DNS_SERVER=	28,	/* ipaddr */
+  PW_MS_SECONDARY_DNS_SERVER=	29,	/* ipaddr */
+  PW_MS_PRIMARY_NBNS_SERVER=	30,	/* ipaddr */
+  PW_MS_SECONDARY_NBNS_SERVER=	31,	/* ipaddr */
+};
+
+/* Vendor RADIUS attribute-value pairs for Roaring Penguin: Bandwidth bit rate limits */
+enum rc_vendor_attr_roaringpenguin {
+  PW_RP_UPSTREAM_LIMIT        =1,  /* integer */
+  PW_RP_DOWNSTREAM_LIMIT      =2,  /* integer */
+};
+
 /* PROHIBIT PROTOCOL */
 #define PW_DUMB			0	//!< 1 and 2 are defined in FRAMED PROTOCOLS.
 #define PW_AUTH_ONLY		3
@@ -433,7 +456,8 @@ typedef enum rc_send_status {
 	ERROR_RC=-1,
 	OK_RC=0,
 	TIMEOUT_RC=1,
-	REJECT_RC=2
+	REJECT_RC=2,
+	CHALLENGE_RC=3
 } rc_send_status;
 
 
@@ -485,8 +509,6 @@ typedef struct rc_aaa_ctx_st RC_AAA_CTX;
 
 /** @} */
 
-__BEGIN_DECLS
-
 /*!\mainpage
  * \section Introduction
  *
@@ -509,8 +531,8 @@ __BEGIN_DECLS
  * parsing the file, simplifying application configuration and administration.
  *
  * Alternative operation without a configuration file is also possible, see
- * rc_add_config().
- * 
+ * rc_add_config() and rc_apply_config().
+ *
  * Check radexample.c for a functional example.
  *
  */
@@ -542,7 +564,9 @@ int rc_avpair_assign (VALUE_PAIR *vp, void const *pval, int len);
 VALUE_PAIR *rc_avpair_new (rc_handle const *rh, int attrid, void const *pval, int len, int vendorpec);
 VALUE_PAIR *rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 			  int length, int vendorpec);
+void rc_avpair_remove (VALUE_PAIR **list, int attrid, int vendorpec);
 VALUE_PAIR *rc_avpair_get (VALUE_PAIR *vp, int attrid, int vendorpec);
+VALUE_PAIR *rc_avpair_copy(VALUE_PAIR *p);
 void rc_avpair_insert(VALUE_PAIR **a, VALUE_PAIR *p, VALUE_PAIR *b);
 void rc_avpair_free (VALUE_PAIR *pair);
 int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first_pair);
@@ -559,7 +583,8 @@ void rc_avpair_get_attr (VALUE_PAIR *vp, unsigned *type, unsigned *id);
 
 void rc_buildreq(rc_handle const *rh, SEND_DATA *data, int code, char *server, unsigned short port,
 		 char *secret, int timeout, int retries);
-int rc_auth(rc_handle *rh, uint32_t client_port, VALUE_PAIR *send, VALUE_PAIR **received, char *msg);
+int rc_auth(rc_handle *rh, uint32_t client_port, VALUE_PAIR *send,
+            VALUE_PAIR **received, char *msg);
 int rc_auth_proxy(rc_handle *rh, VALUE_PAIR *send, VALUE_PAIR **received, char *msg);
 int rc_acct(rc_handle *rh, uint32_t client_port, VALUE_PAIR *send);
 int rc_acct_proxy(rc_handle *rh, VALUE_PAIR *send);
@@ -567,8 +592,13 @@ int rc_check(rc_handle *rh, char *host, char *secret, unsigned short port, char
 
 int rc_aaa(rc_handle *rh, uint32_t client_port, VALUE_PAIR *send, VALUE_PAIR **received,
 	   char *msg, int add_nas_port, rc_standard_codes request_type);
-int rc_aaa_ctx(rc_handle *rh, RC_AAA_CTX **ctx, uint32_t client_port, VALUE_PAIR *send, VALUE_PAIR **received,
-	        char *msg, int add_nas_port, rc_standard_codes request_type);
+int rc_aaa_ctx(rc_handle *rh, RC_AAA_CTX **ctx, uint32_t client_port, VALUE_PAIR *send,
+               VALUE_PAIR **received,
+               char *msg, int add_nas_port, rc_standard_codes request_type);
+int rc_aaa_ctx_server(rc_handle *rh, RC_AAA_CTX **ctx, SERVER *aaaserver,
+                      rc_type type, uint32_t client_port,
+                      VALUE_PAIR *send, VALUE_PAIR **received,
+                      char *msg, int add_nas_port, rc_standard_codes request_type);
 
 /* config.c */
 
@@ -579,6 +609,7 @@ char *rc_conf_str(rc_handle const *rh, char const *optname);
 int rc_conf_int(rc_handle const *rh, char const *optname);
 SERVER *rc_conf_srv(rc_handle const *rh, char const *optname);
 int rc_test_config(rc_handle *rh, char const *filename);
+int rc_apply_config(rc_handle *rh);
 int rc_find_server_addr (rc_handle const *rh, char const *server_name,
                          struct addrinfo** info, char *secret, rc_type type);
 void rc_config_free(rc_handle *rh);
@@ -591,6 +622,7 @@ rc_socket_type rc_get_socket_type(rc_handle * rh);
 /* dict.c */
 
 int rc_read_dictionary (rc_handle *rh, char const *filename);
+int rc_read_dictionary_from_buffer (rc_handle *rh, char const *buf, size_t size);
 DICT_ATTR *rc_dict_getattr(rc_handle const *rh, int attribute);
 DICT_ATTR *rc_dict_findattr(rc_handle const *rh, char const *attrname);
 DICT_VALUE *rc_dict_findval(rc_handle const *rh, char const *valname);
@@ -613,6 +645,7 @@ int rc_get_srcaddr(struct sockaddr *lia, const struct sockaddr *ria);
 
 /* log.c */
 
+void rc_setdebug(int debug);
 void rc_openlog(char const *ident);
 /* to provide compatibility with any old applications that may have
  * been using rc_log() */
@@ -628,6 +661,20 @@ void rc_aaa_ctx_free(RC_AAA_CTX *ctx);
 const char *rc_aaa_ctx_get_secret(RC_AAA_CTX *ctx);
 const void *rc_aaa_ctx_get_vector(RC_AAA_CTX *ctx);
 
-__END_DECLS
+/* obsolete functions */
+#define _RADCLI_GCC_VERSION (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__)
+#if !defined RADCLI_INTERNAL_BUILD
+# if _RADCLI_GCC_VERSION >= 30100
+#  define _RADCLI_GCC_ATTR_DEPRECATED __attribute__ ((__deprecated__))
+# endif
+#endif
+char *rc_mksid(void) _RADCLI_GCC_ATTR_DEPRECATED;
+
+
+/* *INDENT-OFF* */
+#ifdef __cplusplus
+}
+#endif
+/* *INDENT-ON* */
 
 #endif /* RADCLI_H */
