@@ -24,7 +24,7 @@
 
 /**
  * @defgroup radcli-api Main API
- * @brief Main API Functions 
+ * @brief Main API Functions
  *
  * @{
  */
@@ -58,6 +58,48 @@ VALUE_PAIR *rc_avpair_add (rc_handle const *rh, VALUE_PAIR **list, int attrid, v
 
 }
 
+/** Removes an attribute-value pair from the given list
+ *
+ * See rc_avpair_assign() for the format of the data.
+ *
+ * @param list a VALUE_PAIR array of values
+ * @param attrid The attribute of the pair to remove (e.g., PW_USER_NAME).
+ * @param vendorpec The vendor ID in case of a vendor specific value - 0 otherwise.
+ */
+void rc_avpair_remove (VALUE_PAIR **list, int attrid, int vendorpec)
+{
+	VALUE_PAIR *vp, *prev, *tmp;
+	int vattrid;
+
+	if(vendorpec != VENDOR_NONE)
+		vattrid = attrid | (vendorpec << 16);
+	else
+		vattrid = attrid;
+
+	prev = NULL;
+	vp = *list;
+	while(vp != NULL) {
+		if (vp->attribute == vattrid) {
+			/* found */
+			if (prev == NULL) { /* first one */
+				tmp = vp;
+				vp = tmp->next;
+				free(tmp);
+				*list = vp;
+			} else { /* somewhere in the middle */
+				prev->next = vp->next;
+				free(vp);
+			}
+			break;
+		}
+
+		prev = vp;
+		vp = vp->next;
+	};
+
+	return;
+}
+
 /** Iterates through the attribute-value pairs
  *
  * The attribute-value are organized in a linked-list, and this
@@ -152,11 +194,16 @@ VALUE_PAIR *rc_avpair_new (rc_handle const *rh, int attrid, void const *pval, in
 {
 	VALUE_PAIR     *vp = NULL;
 	DICT_ATTR      *pda;
+	int vattrid;
 
-	attrid = attrid | (vendorpec << 16);
-	if ((pda = rc_dict_getattr (rh, attrid)) == NULL)
+	if(vendorpec != VENDOR_NONE) {
+		vattrid = attrid | (vendorpec << 16);
+	} else {
+		vattrid = attrid;
+	}
+	if ((pda = rc_dict_getattr (rh, vattrid)) == NULL)
 	{
-		rc_log(LOG_ERR,"rc_avpair_new: no attribute %d in dictionary", attrid);
+		rc_log(LOG_ERR,"rc_avpair_new: no attribute %d/%u in dictionary", vendorpec, attrid);
 		return NULL;
 	}
 	if (vendorpec != 0 && rc_dict_getvend(rh, vendorpec) == NULL)
@@ -167,7 +214,7 @@ VALUE_PAIR *rc_avpair_new (rc_handle const *rh, int attrid, void const *pval, in
 	if ((vp = malloc (sizeof (VALUE_PAIR))) != NULL)
 	{
 		strlcpy (vp->name, pda->name, sizeof (vp->name));
-		vp->attribute = attrid;
+		vp->attribute = vattrid;
 		vp->next = NULL;
 		vp->type = pda->type;
 		if (rc_avpair_assign (vp, pval, len) == 0)
@@ -248,8 +295,6 @@ VALUE_PAIR *rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char c
 	if (length != attrlen) {
 		pair = rc_avpair_gen(rh, pair, ptr + attrlen, length - attrlen,
 		    vendorpec);
-		if (pair == NULL)
-			return NULL;
 	}
 
 	/* Actual processing */
@@ -308,7 +353,7 @@ VALUE_PAIR *rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char c
 	/* Insert this new pair at the beginning of the list */
 	rpair->next = pair;
 	pair = rpair;
-	strcpy(pair->name, attr->name);
+	strlcpy(pair->name, attr->name, sizeof(pair->name));
 	pair->attribute = attr->value;
 	pair->type = attr->type;
 
@@ -394,6 +439,38 @@ VALUE_PAIR *rc_avpair_get (VALUE_PAIR *vp, int attrid, int vendorpec)
 	return vp;
 }
 
+/** Return a copy of the existing list "p" ala strdup().
+ *
+ * @param p a pointer to a VALUE_PAIR structure.
+ * @return the copy of "p".
+ *
+ */
+VALUE_PAIR *rc_avpair_copy(VALUE_PAIR *p)
+{
+	VALUE_PAIR *vp, *fp = NULL, *lp = NULL;
+
+	while (p) {
+		vp = malloc(sizeof(VALUE_PAIR));
+		if (!vp) {
+			while(fp) { /* free allocated memory */
+				vp = fp;
+				fp = fp->next;
+				free(vp);
+			}
+			return NULL;
+		}
+		*vp = *p;
+		if (!fp)
+			fp = vp;
+		if (lp)
+			lp->next = vp;
+		lp = vp;
+		p = p->next;
+	}
+
+	return fp;
+}
+
 /** Insert a VALUE_PAIR into a list
  *
  * Given the address of an existing list "a" and a pointer to an entry "p" in that list, add the value pair "b" to
@@ -443,8 +520,10 @@ void rc_avpair_insert(VALUE_PAIR **a, VALUE_PAIR *p, VALUE_PAIR *b)
 		}
 	}
 
-	b->next = this_node->next;
-	this_node->next = b;
+	if (this_node) {
+		b->next = this_node->next;
+		this_node->next = b;
+	}
 
 	return;
 }
@@ -555,7 +634,7 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 			mode = PARSE_MODE_EQUAL;
 			break;
 
-		    case PARSE_MODE_EQUAL:		/* Equal sign */
+			case PARSE_MODE_EQUAL:		/* Equal sign */
 			if (*buffer == '=')
 			{
 				mode = PARSE_MODE_VALUE;
@@ -572,7 +651,7 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 			}
 			break;
 
-		    case PARSE_MODE_VALUE:		/* Value */
+			case PARSE_MODE_VALUE:		/* Value */
 			rc_fieldcpy (valstr, &buffer, " \t\n,", sizeof(valstr));
 
 			if ((pair = malloc (sizeof (VALUE_PAIR))) == NULL)
@@ -591,12 +670,12 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 			switch (pair->type)
 			{
 
-			    case PW_TYPE_STRING:
+				case PW_TYPE_STRING:
 				strcpy (pair->strvalue, valstr);
 				pair->lvalue = (uint32_t)strlen(valstr);
 				break;
 
-			    case PW_TYPE_INTEGER:
+				case PW_TYPE_INTEGER:
 				if (isdigit (*valstr))
 				{
 					pair->lvalue = atoi (valstr);
@@ -621,43 +700,43 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 				}
 				break;
 
-			    case PW_TYPE_IPADDR:
-			    	if (inet_pton(AF_INET, valstr, &pair->lvalue) == 0) {
-			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv4 address %s", valstr);
-			    		free(pair);
-			    		return -1;
-			    	}
+				case PW_TYPE_IPADDR:
+					if (inet_pton(AF_INET, valstr, &pair->lvalue) == 0) {
+						rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv4 address %s", valstr);
+						free(pair);
+						return -1;
+					}
 
-                                pair->lvalue = ntohl(pair->lvalue);
+					pair->lvalue = ntohl(pair->lvalue);
 				break;
 
-			    case PW_TYPE_IPV6ADDR:
-			    	if (inet_pton(AF_INET6, valstr, pair->strvalue) == 0) {
-			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 address %s", valstr);
-			    		free(pair);
-			    		return -1;
-			    	}
-				pair->lvalue = 16;
+				case PW_TYPE_IPV6ADDR:
+					if (inet_pton(AF_INET6, valstr, pair->strvalue) == 0) {
+						rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 address %s", valstr);
+						free(pair);
+						return -1;
+					}
+					pair->lvalue = 16;
 				break;
 
-			    case PW_TYPE_IPV6PREFIX:
-			    	p = strchr(valstr, '/');
-			    	if (p == NULL) {
-			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 prefix %s", valstr);
-			    		free(pair);
-			    		return -1;
-			    	}
-			    	*p = 0;
-			    	p++;
-			    	pair->strvalue[0] = 0;
-			    	pair->strvalue[1] = atoi(p);
+				case PW_TYPE_IPV6PREFIX:
+					p = strchr(valstr, '/');
+					if (p == NULL) {
+						rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 prefix %s", valstr);
+						free(pair);
+						return -1;
+					}
+					*p = 0;
+					p++;
+					pair->strvalue[0] = 0;
+					pair->strvalue[1] = atoi(p);
 
-			    	if (inet_pton(AF_INET6, valstr, pair->strvalue+2) == 0) {
-			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 prefix %s", valstr);
-			    		free(pair);
-			    		return -1;
-			    	}
-				pair->lvalue = 2+16;
+					if (inet_pton(AF_INET6, valstr, pair->strvalue+2) == 0) {
+						rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 prefix %s", valstr);
+						free(pair);
+						return -1;
+					}
+					pair->lvalue = 2+16;
 				break;
 
 			    case PW_TYPE_DATE:
@@ -674,7 +753,7 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 #endif	/* TIMELOCAL */
 				break;
 
-			    default:
+				default:
 				rc_log(LOG_ERR, "rc_avpair_parse: unknown attribute type %d", pair->type);
 				if (*first_pair) {
 					rc_avpair_free(*first_pair);
@@ -729,7 +808,7 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 			mode = PARSE_MODE_NAME;
 			break;
 
-		    default:
+			default:
 			mode = PARSE_MODE_NAME;
 			break;
 		}
@@ -753,6 +832,7 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 	struct in_addr  inad;
 	unsigned char  *ptr;
 	unsigned int    pos;
+	unsigned int    slen;
 
 	*name = *value = '\0';
 
@@ -765,15 +845,17 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 
 	switch (pair->type)
 	{
-	    case PW_TYPE_STRING:
-	    	lv--;
-	    	pos = 0;
+		case PW_TYPE_STRING:
+		lv--;
+		pos = 0;
 		ptr = (unsigned char *) pair->strvalue;
 		if (pair->attribute == PW_DIGEST_ATTRIBUTES) {
-			pair->strvalue[*(ptr + 1)] = '\0';
+			slen = ptr[1] - 2;
 			ptr += 2;
+		} else {
+			slen = pair->lvalue;
 		}
-		while (*ptr != '\0')
+		while (slen-- > 0)
 		{
 			if (!(isprint (*ptr)))
 			{
@@ -802,7 +884,7 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 			value[pos-1] = 0;
 		break;
 
-	    case PW_TYPE_INTEGER:
+		case PW_TYPE_INTEGER:
 		dval = rc_dict_getval (rh, pair->lvalue, pair->name);
 		if (dval != NULL)
 		{
@@ -814,37 +896,37 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 		}
 		break;
 
-	    case PW_TYPE_IPADDR:
+		case PW_TYPE_IPADDR:
 		inad.s_addr = htonl(pair->lvalue);
 		strlcpy (value, inet_ntoa (inad), (size_t) lv);
 		break;
 
-	    case PW_TYPE_IPV6ADDR:
-	    	if (inet_ntop(AF_INET6, pair->strvalue, value, lv) == NULL)
-	    		return -1;
+		case PW_TYPE_IPV6ADDR:
+			if (inet_ntop(AF_INET6, pair->strvalue, value, lv) == NULL)
+				return -1;
 		break;
 
-	    case PW_TYPE_IPV6PREFIX: {
-	    	uint8_t ip[16];
-	    	uint8_t txt[48];
-	    	if (pair->lvalue < 2)
-	    		return -1;
+		case PW_TYPE_IPV6PREFIX: {
+			uint8_t ip[16];
+			uint8_t txt[48];
+			if (pair->lvalue < 2)
+				return -1;
 
-	    	memset(ip, 0, sizeof(ip));
-	    	memcpy(ip, pair->strvalue+2, pair->lvalue-2);
+			memset(ip, 0, sizeof(ip));
+			memcpy(ip, pair->strvalue+2, pair->lvalue-2);
 
-	    	if (inet_ntop(AF_INET6, ip, (void*)txt, sizeof(txt)) == NULL)
-	    		return -1;
+			if (inet_ntop(AF_INET6, ip, (void*)txt, sizeof(txt)) == NULL)
+				return -1;
 		snprintf(value, lv, "%s/%u", txt, (unsigned)pair->strvalue[1]);
 
 		break;
-	    }
-	    case PW_TYPE_DATE:
+		}
+		case PW_TYPE_DATE:
 		strftime (value, lv, "%m/%d/%y %H:%M:%S",
 			  gmtime ((time_t *) & pair->lvalue));
 		break;
 
-	    default:
+		default:
 		rc_log(LOG_ERR, "rc_avpair_tostr: unknown attribute type %d", pair->type);
 		return -1;
 		break;
@@ -873,7 +955,7 @@ char *rc_avpair_log(rc_handle const *rh, VALUE_PAIR *pair, char *buf, size_t buf
 	for (vp = pair; vp != NULL; vp = vp->next) {
 		if (rc_avpair_tostr(rh, vp, name, sizeof(name), value,
 		    sizeof(value)) == -1)
-		        return NULL;
+				return NULL;
 		nlen = len + 32 + 3 + strlen(value) + 2 + 2;
 		if(nlen<buf_len-1) {
 			sprintf(buf + len, "%-32s = '%s'\n", name, value);
@@ -897,8 +979,8 @@ int rc_avpair_get_uint32 (VALUE_PAIR *vp, uint32_t *res)
 {
 	if (vp->type == PW_TYPE_DATE || vp->type == PW_TYPE_IPADDR ||
 	    vp->type == PW_TYPE_INTEGER) {
-	    	if (res)
-		    *res = vp->lvalue;
+			if (res)
+				*res = vp->lvalue;
 		return 0;
 	} else {
 		return -1;
@@ -920,17 +1002,17 @@ int rc_avpair_get_in6 (VALUE_PAIR *vp, struct in6_addr *res, unsigned *prefix)
 		memcpy(res, vp->strvalue, 16);
 		return 0;
 	} else if (vp->type == PW_TYPE_IPV6PREFIX) {
-	    	if (vp->lvalue < 2 || vp->lvalue > 18)
-	    		return -1;
+			if (vp->lvalue < 2 || vp->lvalue > 18)
+				return -1;
 
 		if (res) {
-		    	memset(res, 0, 16);
-		    	memcpy(res, vp->strvalue+2, vp->lvalue-2);
+			memset(res, 0, 16);
+			memcpy(res, vp->strvalue+2, vp->lvalue-2);
 		}
 
 		if (prefix)
-		    	*prefix = (unsigned char)vp->strvalue[1];
-	    	return 0;
+			*prefix = (unsigned char)vp->strvalue[1];
+		return 0;
 	}
 
 	return -1;
@@ -949,9 +1031,9 @@ int rc_avpair_get_in6 (VALUE_PAIR *vp, struct in6_addr *res, unsigned *prefix)
 int rc_avpair_get_raw (VALUE_PAIR *vp, char **res, unsigned *res_size)
 {
 	if (vp->type == PW_TYPE_STRING || vp->type == PW_TYPE_IPV6ADDR ||
-	    vp->type == PW_TYPE_IPV6PREFIX) {
-	    	if (res)
-	    		*res = vp->strvalue;
+		vp->type == PW_TYPE_IPV6PREFIX) {
+			if (res)
+				*res = vp->strvalue;
 		if (res_size)
 			*res_size = vp->lvalue;
 		return 0;
@@ -975,3 +1057,9 @@ void rc_avpair_get_attr (VALUE_PAIR *vp, unsigned *type, unsigned *id)
 }
 
 /** @} */
+/*
+ * Local Variables:
+ * c-basic-offset:8
+ * c-style: whitesmith
+ * End:
+ */
