@@ -18,6 +18,17 @@
 #include <poll.h>
 #include "util.h"
 #include "rc-md5.h"
+#include "rc-hmac.h"
+
+#if defined(HAVE_GNUTLS)
+# include <gnutls/gnutls.h>
+# include <gnutls/crypto.h>
+#endif
+
+#if defined(__linux__)
+#include <linux/in6.h>
+#endif
+
 
 #define SCLOSE(fd) if (sfuncs->close_fd) sfuncs->close_fd(fd)
 
@@ -27,7 +38,7 @@ static int rc_check_reply(AUTH_HDR *, int, char const *, unsigned char const *,
 
 /**
  * @defgroup radcli-api Main API
- * @brief Main API Functions 
+ * @brief Main API Functions
  *
  * @{
  */
@@ -110,36 +121,6 @@ static int rc_pack_list(VALUE_PAIR * vp, char *secret, AUTH_HDR * auth)
 			total_length += padded_length + 2;
 
 			break;
-#if 0
-		case PW_CHAP_PASSWORD:
-
-			*buf++ = CHAP_VALUE_LENGTH + 2;
-			if (vsa_length_ptr != NULL)
-				*vsa_length_ptr += CHAP_VALUE_LENGTH + 2;
-
-			/* Encrypt the Password */
-			length = vp->lvalue;
-			if (length > CHAP_VALUE_LENGTH) {
-				length = CHAP_VALUE_LENGTH;
-			}
-			memset((char *)passbuf, '\0', CHAP_VALUE_LENGTH);
-			memcpy((char *)passbuf, vp->strvalue, (size_t) length);
-
-			/* Calculate the MD5 Digest */
-			secretlen = strlen(secret);
-			strcpy((char *)md5buf, secret);
-			memcpy((char *)md5buf + secretlen, (char *)auth->vector,
-			       AUTH_VECTOR_LEN);
-			rc_md5_calc(buf, md5buf, secretlen + AUTH_VECTOR_LEN);
-
-			/* Xor the password into the MD5 digest */
-			for (i = 0; i < CHAP_VALUE_LENGTH; i++) {
-				*buf++ ^= passbuf[i];
-			}
-			total_length += CHAP_VALUE_LENGTH + 2;
-
-			break;
-#endif
 		default:
 			switch (vp->type) {
 			case PW_TYPE_STRING:
@@ -367,7 +348,11 @@ static void rc_random_vector(unsigned char *vector)
 {
 	int randno;
 	int i;
-#if defined(HAVE_GETENTROPY)
+#if defined(HAVE_GNUTLS)
+	if (gnutls_rnd(GNUTLS_RND_NONCE, vector, AUTH_VECTOR_LEN) >= 0) {
+		return;
+	}
+#elif defined(HAVE_GETENTROPY)
 	if (getentropy(vector, AUTH_VECTOR_LEN) >= 0) {
 		return;
 	}			/* else fall through */
@@ -398,8 +383,8 @@ static void rc_random_vector(unsigned char *vector)
 		close(fd);
 		return;
 	}			/* else fall through */
-#endif
  fallback:
+#endif
 	for (i = 0; i < AUTH_VECTOR_LEN;) {
 		randno = random();
 		memcpy((char *)vector, (char *)&randno, sizeof(int));
@@ -412,6 +397,38 @@ static void rc_random_vector(unsigned char *vector)
 
 /** @} */
 
+
+/** Add a Message-Authenticator attribute to a message. This is mandatory,
+ *  for example, when sending a message containing an EAP-Message
+ *  attribute.
+ *
+ * @param rh - A handle to parsed configuration
+ * @param secret - The server's secret string
+ * @param auth - Pointer to the AUTH_HDR structure
+ * @param total_length - Total packet length before Message Authenticator
+ *                is added.
+ *
+ * @return Total packet length after Message Authenticator is added.
+ */
+static int add_msg_auth_attr(rc_handle * rh, char * secret,
+			AUTH_HDR *auth, int total_length)
+{
+	size_t secretlen = strlen(secret);
+	uint8_t *msg_auth = (uint8_t *)auth + total_length;
+	msg_auth[0] = PW_MESSAGE_AUTHENTICATOR;
+	msg_auth[1] = 18;
+	memset(&msg_auth[2], 0, MD5_DIGEST_SIZE);
+	total_length += 18;
+	auth->length = htons((unsigned short)total_length);
+
+	/* Calulate HMAC-MD5 [RFC2104] hash */
+	uint8_t digest[MD5_DIGEST_SIZE];
+	rc_hmac_md5((uint8_t *)auth, (size_t)total_length, (uint8_t *)secret, secretlen, digest);
+	memcpy(&msg_auth[2], digest, MD5_DIGEST_SIZE);
+
+	return total_length;
+}
+
 /** Sends a request to a RADIUS server and waits for the reply
  *
  * @param rh a handle to parsed configuration
@@ -420,8 +437,9 @@ static void rc_random_vector(unsigned char *vector)
  * @param msg must be an array of %PW_MAX_MSG_SIZE or NULL; will contain the concatenation of
  *	any %PW_REPLY_MESSAGE received.
  * @param type must be %AUTH or %ACCT
- * @return OK_RC (0) on success, TIMEOUT_RC on timeout REJECT_RC on acess reject, or negative
- *	on failure as return value.
+ * @return OK_RC (0) on success, CHALLENGE_RC when an Access-Challenge
+ *  response is received, TIMEOUT_RC on timeout REJECT_RC on access reject,
+ *  or negative on failure as return value.
  */
 int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 		       char *msg, rc_type type)
@@ -444,25 +462,37 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 	uint8_t recv_buffer[BUFFER_LEN];
 	uint8_t send_buffer[BUFFER_LEN];
 	uint8_t *attr;
+	uint16_t tlen;
 	int retries;
 	VALUE_PAIR *vp;
 	struct pollfd pfd;
 	double start_time, timeout;
 	struct sockaddr_storage *ss_set = NULL;
 	char *server_type = "auth";
+	char *ns = NULL;
+	int ns_def_hdl = 0;
 
 	server_name = data->server;
 	if (server_name == NULL || server_name[0] == '\0')
 		return ERROR_RC;
 
+	ns = rc_conf_str(rh, "namespace"); /* Check for namespace config */
+	if (ns != NULL) {
+		if(-1 == rc_set_netns(ns, &ns_def_hdl)) {
+			rc_log(LOG_ERR, "rc_send_server: namespace %s set failed", ns);
+			return ERROR_RC;
+		}
+	}
 	if ((vp = rc_avpair_get(data->send_pairs, PW_SERVICE_TYPE, 0)) &&
 	    (vp->lvalue == PW_ADMINISTRATIVE)) {
 		strcpy(secret, MGMT_POLL_SECRET);
 		auth_addr =
 		    rc_getaddrinfo(server_name,
 				   type == AUTH ? PW_AI_AUTH : PW_AI_ACCT);
-		if (auth_addr == NULL)
-			return ERROR_RC;
+		if (auth_addr == NULL) {
+			result = ERROR_RC;
+			goto exit_error;
+		}    
 	} else {
 		if (data->secret != NULL) {
 			strlcpy(secret, data->secret, MAX_SECRET_LENGTH);
@@ -476,7 +506,8 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 			rc_log(LOG_ERR,
 			       "rc_send_server: unable to find server: %s",
 			       server_name);
-			return ERROR_RC;
+			result = ERROR_RC;
+			goto exit_error;
 		}
 		/*} */
 	}
@@ -492,7 +523,8 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 	if (sfuncs->lock) {
 		if (sfuncs->lock(sfuncs->ptr) != 0) {
 			rc_log(LOG_ERR, "%s: lock error", __func__);
-			return ERROR_RC;
+			result = ERROR_RC;
+			goto exit_error;
 		}
 	}
 
@@ -529,6 +561,35 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 		}
 	}
 
+	if(our_sockaddr.ss_family  == AF_INET6) {
+		/* Check for IPv6 non-temporary address support */
+		char *non_temp_addr = rc_conf_str(rh, "use-public-addr");
+		if (non_temp_addr && (strcasecmp(non_temp_addr, "true") == 0)) {
+#if defined(__linux__)
+			int sock_opt = IPV6_PREFER_SRC_PUBLIC;
+			if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_ADDR_PREFERENCES, 
+					&sock_opt, sizeof(sock_opt)) != 0) {
+				rc_log(LOG_ERR, "rc_send_server: setsockopt: %s", 
+					strerror(errno));
+				result = ERROR_RC;
+				goto cleanup;
+			}
+#elif defined(BSD) || defined(__APPLE__)
+			int sock_opt = 0;
+			if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_PREFER_TEMPADDR,
+				&sock_opt, sizeof(sock_opt)) != 0) {
+				rc_log(LOG_ERR, "rc_send_server: setsockopt: %s", 
+					strerror(errno));
+				result = ERROR_RC;
+				goto cleanup;
+			}
+#else
+			rc_log(LOG_INFO, "rc_send_server: Usage of non-temporary IPv6"
+					" address is not supported in this system");
+#endif       
+		}
+	}
+
 	retry_max = data->retries;	/* Max. numbers to try for reply */
 	retries = 0;		/* Init retry cnt for blocking call */
 
@@ -594,7 +655,8 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 		total_length =
 		    rc_pack_list(data->send_pairs, secret, auth) + AUTH_HDR_LEN;
 
-		auth->length = htons((unsigned short)total_length);
+		tlen = htons((unsigned short)total_length);
+		memcpy(&auth->length, &tlen, sizeof(uint16_t));
 
 		memset((char *)auth->vector, 0, AUTH_VECTOR_LEN);
 		secretlen = strlen(secret);
@@ -610,6 +672,10 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 		    rc_pack_list(data->send_pairs, secret, auth) + AUTH_HDR_LEN;
 
 		auth->length = htons((unsigned short)total_length);
+
+		/* If EAP message we MUST add a Message-Authenticator attribute */
+		if (rc_avpair_get(data->send_pairs, PW_EAP_MESSAGE, 0) != NULL)
+		    total_length = add_msg_auth_attr(rh, secret, auth, total_length);
 	}
 
 	if (radcli_debug) {
@@ -646,9 +712,9 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 		pfd.fd = sockfd;
 		pfd.events = POLLIN;
 		pfd.revents = 0;
-		start_time = rc_getctime();
+		start_time = rc_getmtime();
 		for (timeout = data->timeout; timeout > 0;
-		     timeout -= rc_getctime() - start_time) {
+		     timeout -= rc_getmtime() - start_time) {
 			result = poll(&pfd, 1, timeout * 1000);
 			if (result != -1 || errno != EINTR)
 				break;
@@ -809,15 +875,25 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 		}
 	}
 
-	if ((recv_auth->code == PW_ACCESS_ACCEPT) ||
-	    (recv_auth->code == PW_PASSWORD_ACK) ||
-	    (recv_auth->code == PW_ACCOUNTING_RESPONSE)) {
+	switch (recv_auth->code) {
+	case PW_ACCESS_ACCEPT:
+	case PW_PASSWORD_ACK:
+	case PW_ACCOUNTING_RESPONSE:
 		result = OK_RC;
-	} else if ((recv_auth->code == PW_ACCESS_REJECT) ||
-		   (recv_auth->code == PW_PASSWORD_REJECT)) {
+		break;
+
+	case PW_ACCESS_REJECT:
+	case PW_PASSWORD_REJECT:
 		result = REJECT_RC;
-	} else {
-		rc_log(LOG_ERR, "rc_send_server: received RADIUS server response neither ACCEPT nor REJECT, invalid");
+		break;
+
+	case PW_ACCESS_CHALLENGE:
+		result = CHALLENGE_RC;
+		break;
+
+	default:
+		rc_log(LOG_ERR, "rc_send_server: received RADIUS server response neither ACCEPT nor REJECT, code=%d is invalid",
+		       recv_auth->code);
 		result = BADRESP_RC;
 	}
 
@@ -830,6 +906,13 @@ int rc_send_server_ctx(rc_handle * rh, RC_AAA_CTX ** ctx, SEND_DATA * data,
 			rc_log(LOG_ERR, "%s: unlock error", __func__);
 		}
 	}
+ exit_error:
+	if (ns != NULL) {
+		if(-1 == rc_reset_netns(&ns_def_hdl)) {
+			rc_log(LOG_ERR, "rc_send_server: namespace %s reset failed", ns);
+			result = ERROR_RC;
+		}
+	}
 
 	return result;
 }
