@@ -35,6 +35,12 @@
  * @defgroup tls-api TLS/DTLS API
  * @brief TLS and DTLS related functions
  *
+ * Note that, that API is for improving TLS and DTLS support
+ * in an application. Applications are not required to use this
+ * API to support them. TLS and DTLS support can be enabled by
+ * the administrator transparently from the radiusclient configuration
+ * file.
+ *
  * @{
  */
 
@@ -419,9 +425,12 @@ static void restart_session(rc_handle *rh, tls_st *st)
 }
 
 /** Returns the file descriptor of the TLS/DTLS session
+ *
+ * This can also be used as a test for the application to see
+ * whether TLS or DTLS are in use.
  *
  * @param rh a handle to parsed configuration
- * @return the file descriptor used by the TLS session
+ * @return the file descriptor used by the TLS session, or -1 on error
  */
 int rc_tls_fd(rc_handle * rh)
 {
@@ -488,13 +497,27 @@ int rc_check_tls(rc_handle * rh)
 void rc_deinit_tls(rc_handle * rh)
 {
 	tls_st *st = rh->so.ptr;
+	char *ns = NULL;
+	int ns_def_hdl = 0;
+
 	if (st) {
+		ns = rc_conf_str(rh, "namespace"); /* Check for namespace config */
+		if (ns != NULL) {
+			if(-1 == rc_set_netns(ns, &ns_def_hdl)) {
+				rc_log(LOG_ERR, "rc_send_server: namespace %s set failed", ns);
+				return;
+			}
+		}
 		if (st->ctx.init != 0)
 			deinit_session(&st->ctx);
 		if (st->x509_cred)
 			gnutls_certificate_free_credentials(st->x509_cred);
 		if (st->psk_cred)
 			gnutls_psk_free_client_credentials(st->psk_cred);
+		if (ns != NULL) {
+			if(-1 == rc_reset_netns(&ns_def_hdl))
+			rc_log(LOG_ERR, "rc_send_server: namespace %s reset failed", ns);
+		}
 	}
 	free(st);
 }
@@ -519,9 +542,19 @@ int rc_init_tls(rc_handle * rh, unsigned flags)
 	SERVER *authservers;
 	char hostname[256];	/* server's hostname */
 	unsigned port;		/* server's port */
+	char *ns = NULL;
+	int ns_def_hdl = 0;
 
 	memset(&rh->so, 0, sizeof(rh->so));
 
+	ns = rc_conf_str(rh, "namespace"); /* Check for namespace config */
+	if (ns != NULL) {
+		if(-1 == rc_set_netns(ns, &ns_def_hdl)) {
+			rc_log(LOG_ERR, "rc_send_server: namespace %s set failed", ns);
+			return -1;
+		}
+	}
+
 	if (flags & SEC_FLAG_DTLS) {
 		rh->so_type = RC_SOCKET_DTLS;
 		rh->so.static_secret = DEFAULT_DTLS_SECRET;
@@ -561,8 +594,8 @@ int rc_init_tls(rc_handle * rh, unsigned flags)
 			if (ret < 0) {
 				ret = -1;
 				rc_log(LOG_ERR,
-				       "%s: error in setting X.509 trust file: %s",
-				       __func__, gnutls_strerror(ret));
+				       "%s: error in setting X.509 trust file: %s: %s",
+				       __func__, gnutls_strerror(ret), ca_file);
 				goto cleanup;
 			}
 		}
@@ -576,8 +609,8 @@ int rc_init_tls(rc_handle * rh, unsigned flags)
 			if (ret < 0) {
 				ret = -1;
 				rc_log(LOG_ERR,
-				       "%s: error in setting X.509 cert and key file: %s",
-				       __func__, gnutls_strerror(ret));
+				       "%s: error in setting X.509 cert and key files: %s: %s - %s",
+				       __func__, gnutls_strerror(ret), cert_file, key_file);
 				goto cleanup;
 			}
 		}
@@ -676,6 +709,12 @@ int rc_init_tls(rc_handle * rh, unsigned flags)
 	rh->so.recvfrom = tls_recvfrom;
 	rh->so.lock = tls_lock;
 	rh->so.unlock = tls_unlock;
+	if (ns != NULL) {
+		if(-1 == rc_reset_netns(&ns_def_hdl)) {
+			rc_log(LOG_ERR, "rc_send_server: namespace %s reset failed", ns);
+			goto cleanup;
+		}
+	}
 	return 0;
  cleanup:
 	if (st) {
@@ -687,6 +726,10 @@ int rc_init_tls(rc_handle * rh, unsigned flags)
 			gnutls_psk_free_client_credentials(st->psk_cred);
 	}
 	free(st);
+	if (ns != NULL) {
+		if(-1 == rc_reset_netns(&ns_def_hdl))
+		rc_log(LOG_ERR, "rc_send_server: namespace %s reset failed", ns);
+	}
 	return ret;
 }
 
